#!/usr/bin/env python3
#
# Copyright 2019. Clumio, Inc.

import boto3
import click
import pandas as pd
from botocore.exceptions import ClientError
import openpyxl
import datetime
import pprint as pp


@click.group()
def cli():
    '''

    Command line tools by Clumio, inc.

    Maintained by Luke McCleary | Luke@clumio.com

    '''
    pass


@cli.command()
@click.option('-id', '--key-id', prompt=True)
@click.option('-sk', '--secret-key', prompt=True, hide_input=True)
@click.option('-o', '--outfile', default='Clumio_Calipers.xlsx')
@click.option('-g', '--geo', multiple=True, required=True, default=['am'],
              type=click.Choice(['eu', 'am', 'apac'], case_sensitive=True))
def report(key_id, secret_key, geo, outfile):
    '''
    Generates an excel report on EC2, EBS, RDS, etc assets. Defaults to
    North and South American regions, non-govcloud. Other regions,
    or for multiple regions, us the --geo or -g flags.

    Example 1: Let the script prompt for required inputs: \n
        Python3 -m calipers report

    Example 2: Specify output file name and a geo: \n
        Python3 -m calipers report -o filename.xlsx -g eu

    Example 3: Specify output file name and multiple geos: \n
        Python3 -m calipers report -o filename.xlsx -g eu -g am

    Americas = am \n
    Europe = eu \n
    Asia = apac \n

    '''
    americas = ['us-east-1', 'us-east-2', 'us-west-1',
                'us-west-2', 'ca-central-1', 'sa-east-1']
    europe = ['eu-central-1', 'eu-west-1', 'me-south-1',
              'eu-west-2', 'eu-west-3', 'eu-north-1']
    asiapac = ['ap-east-1', 'ap-south-1',
               'ap-northeast-3', 'ap-northeast-2', 'ap-southeast-1',
               'ap-southeast-2', 'ap-northeast-1']

    regions = []
    if 'am' in geo:
        for i in americas:
            regions.append(i)
    if 'eu' in geo:
        for i in europe:
            regions.append(i)
    if 'apac' in geo:
        for i in asiapac:
            regions.append(i)

    print('\nReporting on the following AWS Regions' + str(regions) + '\n')

    '''initialize dictionaries'''
    ec2 = []
    ebs_vols_active = []
    ebs_vols_orph = []
    ebs_snaps = []
    orphan_ebs_snaps = []
    rds_inst = []
    rds_clust = []
    # dyanmo_dict = []

    for r in regions:
        '''Regional iterative with connection to relevent service'''
        try:
            '''Gather EC2 info and append to dicts'''
            ec2_conn = connect_service(key_id, secret_key, r, 'ec2')
            for i in (ec2_report(ec2_conn, r)):
                ec2.append(i)
            active, orphans = ebs_report(ec2_conn, r)
            for a in active:
                ebs_vols_active.append(a)
            for o in orphans:
                ebs_vols_orph.append(o)
            for i in (ebs_snap_report(ec2_conn, non_orphan_snaps, r)):
                ebs_snaps.append(i)
            for i in orphan_ebs_snap_report(ec2_conn, r):
                orphan_ebs_snaps.append(i)

            '''Gather RDS info and append to dicts'''
            rds_conn = connect_service(key_id, secret_key, r, 'rds')
            for i in (rds_inst_report(rds_conn, r)):
                rds_inst.append(i)
            for i in (rds_clusters_report(rds_conn, r)):
                rds_clust.append(i)

            '''gather DynamoDB info and append to dicts'''
            # dynamo_conn = connect_service(key_id, secret_key, r, 'dynamodb')
            # dynamo = dynamo_report(dynamo_conn, r, dyanmo_dict)

        except Exception as e:
            print('A general error occured returning data from AWS\n' + str(e))

    try:
        '''convert returned dicts to dataframes'''
        ec2_df = dataframer(ec2, 'InstanceId')
        active_ebs_df = dataframer(ebs_vols_active, 'VolumeId')
        orph_ebs_df = dataframer(ebs_vols_orph, 'VolumeId')
        ebs_snap_df = dataframer(ebs_snaps, 'SnapshotId')
        orph_ebs_snap_df = dataframer(orphan_ebs_snaps, 'SnapshotId')
        rds_inst_df = dataframer(rds_inst, 'DB Instance Identifier')
        rds_clust_df = dataframer(rds_clust, 'DB Cluster Identifier')
        # dynamo_df = dataframer(dynamo, '')
    except Exception as e:
        print('Error adding python dictionary to pandas dataframe \n' + str(e))

    print('''
    :::::::::::::::::::::::::::::::::::::::::::::\n
    AWS data collection complete. Tidying up... \n
    :::::::::::::::::::::::::::::::::::::::::::::\n
    ''')

    '''Analyze / aggregate the results'''

    '''EC2 / EBS'''
    ec2_num = len(ec2_df.index)
    attached_ebs_num = len(active_ebs_df.index)
    attached_ebs_prov = active_ebs_df['Size (GB)'].sum()
    orph_ebs_num = len(orph_ebs_df.index)
    orph_ebs_prov = orph_ebs_df['Size (GB)'].sum()
    unencrypted_ebs = ((orph_ebs_df["Encrypted?"].str.count('No')).sum() +
                       (active_ebs_df["Encrypted?"].str.count('No')).sum())

    '''unprotected ebs'''
    unprotected_active = active_ebs_df["Total Snaps"].isin(['0'])
    unprotected_orph = orph_ebs_df["Total Snaps"].isin(['0'])
    protected_ebs = (len(unprotected_active.loc[unprotected_active]) +
                     len(unprotected_orph.loc[unprotected_orph]))
    unprotected_ebs = (attached_ebs_num + orph_ebs_num - protected_ebs)

    '''EBS Snaps'''
    backed_snaps_num = len(ebs_snap_df.index)
    orphaned_snaps_num = len(orph_ebs_snap_df.index)
    old_active = ebs_snap_df[ebs_snap_df['60+ Days Old']]
    old_orph = orph_ebs_snap_df.loc[orph_ebs_snap_df['60+ Days Old']]
    ebs_snaps_60plus = (len(old_active) + len(old_orph))

    '''RDS'''
    rds_clustersGB = rds_clust_df['AllocatedStorage (GB)'].sum()
    rds_instancesGB = rds_inst_df['DB Provisioned Capacity (GB)'].sum()
    provisioned_RDS = rds_instancesGB + rds_clustersGB

    summary = []
    summary.append({
        'EC2': '-',
        'EC2 Instances': ec2_num,
        '-': '-',
        'EBS Volumes': '-',
        'Attached EBS Volumes': attached_ebs_num,
        'Attached EBS Provisioned (GB)': attached_ebs_prov,
        'Orphaned EBS Volumes': orph_ebs_num,
        'Orphaned EBS Provisioned (GB)': orph_ebs_prov,
        'Unencrypted Volumes': unencrypted_ebs,
        '--': '--',
        'EBS Snapshots': '-',
        'Unprotected Volumes (no Snaps)': unprotected_ebs,
        'EBS Backed Snaps': backed_snaps_num,
        'Orphaned EBS Snaps': orphaned_snaps_num,
        'Unencrypted Volumes': unencrypted_ebs,
        'EBS Snaps > 60 Days Old': ebs_snaps_60plus,
        'Estimated EBS Snap Size (GB)': '',
        'Estimated EBS Snap Monthly Costs': '',
        '----': '----',
        'RDS': ' ',
        'Total RDS Provisioned (GB)': provisioned_RDS
    })

    summary_df = pd.DataFrame(summary).transpose()
    summary_df.rename(columns={0: 'Value'}, inplace=True)

    try:
        '''pandas commands to merge reports into excel workbooks by tabs'''
        writer = pd.ExcelWriter(outfile)
        summary_df.to_excel(writer, 'AWS_Summary')
        ec2_df.to_excel(writer, 'Running_Stopped_EC2')
        active_ebs_df.to_excel(writer, 'Attached EBS Volumes')
        orph_ebs_df.to_excel(writer, 'Orphan EBS Volumes')
        ebs_snap_df.to_excel(writer, 'EBS Backed Snaphots')
        orph_ebs_snap_df.to_excel(writer, 'Orphan EBS Snapshots')
        rds_inst_df.to_excel(writer, 'RDS_Instances')
        rds_clust_df.to_excel(writer, 'RDS_Clusters')
        writer.save()

    except Exception as e:
        print('\nError Writing to Excel File\n' + str(e))

    print(summary_df)


def ec2_report(ec2_conn, region):
    '''initialize variables'''
    dict = []
    response = []
    response_dict = []
    ec2_vols = []
    try:
        '''fetch running and stopped ec2 instances with paginator'''
        print('\nFetching EC2 Data from ' + str(region) + '...')
        pag = ec2_conn.get_paginator('describe_instances')
        for page in pag.paginate(Filters=[
                {
                    'Name': 'instance-state-name',
                    'Values': ['running', 'stopped']
                },

        ], PaginationConfig={
                'PageSize': 150
        }
        ):
            for i in page['Reservations']:
                response.append(i)
    except Exception as e:
        print('Failed to fetch EC2 Instances' + str(e))

    '''this line can probs be removed by adding [instances] to array above'''
    for d in range(len(response)):
        for i in range(len(response[d]['Instances'])):
            response_dict.append(response[d]['Instances'][i])

    print('\nFound ' + str(len(response_dict)) + ' Running or Stopped EC2 '
          'Instances in ' + str(region))

    try:
        '''fetch all volumes associated with EC2 Instances'''
        ec2_vols = []
        for i in range(len(response_dict)):
            for v in range(len(response_dict[i]['BlockDeviceMappings'])):
                ec2_vols.append(response_dict[i]['BlockDeviceMappings'][v]
                                ['Ebs']['VolumeId'])

        '''fetch detailed volume info for associated EC2 instances'''
        p = Paginator(ec2_vols, 150)
        vols_dict = []
        for i in p:
            response = ec2_conn.describe_volumes(Filters=[
                {
                    'Name': 'volume-id',
                    'Values': i,
                }
            ])['Volumes']
            for i in response:
                vols_dict.append(i)
    except Exception as e:
        print('Failed to fetch volumes assocated with EC2 instances' + str(e))

    try:
        '''iterate through regions appending data to datastructure'''
        for i in range(len(response_dict)):
            '''store volume IDs in local variable for loop'''
            vols = []
            for v in range(len(response_dict[i]['BlockDeviceMappings'])):
                vols.append(response_dict[i]['BlockDeviceMappings'][v]['Ebs']
                            ['VolumeId'])

            '''add up vol sizes'''
            ebs_agg = 0
            for v in vols:
                for d in range(len(vols_dict)):
                    if v == vols_dict[d]['VolumeId']:
                        ebs_agg = ebs_agg + vols_dict[d]['Size']

            '''write ec2 data to data structure'''
            dict.append({
                'InstanceId': response_dict[i]['InstanceId'],
                'Type': response_dict[i]['InstanceType'],
                'Region': str(region),
                'Power-State': response_dict[i]['State']['Name'],
                'Total EBS Provisioned (GB)': ebs_agg,
                'Attached-Volumes': str(vols)
            })
    except Exception as e:
        print('\nNo EC2 Instances found in ' + str(region) + '\n' + str(e))

    return dict


def ebs_report(ec2_conn, region):
    '''does this still need to be a global var?'''
    global non_orphan_snaps
    global vols_in_region
    response = []
    flat_response = []
    vols_in_region = []
    orph_dict = []
    active_dict = []
    non_orphan_snaps = []

    '''create paginator and build ebs_vols reponse dict'''
    print('\nFetching EBS Data from ' + str(region) + '...')
    pag = ec2_conn.get_paginator('describe_volumes')

    for page in pag.paginate(Filters=[
            {
                'Name': 'status',
                'Values': ['available', 'in-use']
            }
        ],
        PaginationConfig={
            'PageSize': 150
    }
    ):
        for i in page['Volumes']:
            response.append(i)
    for d in range(len(response)):
        flat_response.append(response[d])

    print('\nFound ' + str(len(flat_response)) + ' Available or In-Use '
          'EBS volumes in ' + str(region))

    '''Store list of active volumes for subsequent snapshot query'''
    try:
        for i in range(len(flat_response)):
            vols_in_region.append(flat_response[i]['VolumeId'])
    except Exception as e:
        print('Failed to add volumes IDs to regional list\n' + str(e))

    '''filter for only non-orphaned volumes'''
    active_ebs = []
    try:
        for r in range(len(flat_response)):
            if flat_response[r]['State'] == 'in-use':
                active_ebs.append(flat_response[r])
    except Exception as e:
        print('Failed to filter for in-use EBS volumes ' + str(e))

    '''filter for only orphaned volumes'''
    orph_ebs = []
    try:
        for r in range(len(flat_response)):
            if flat_response[r]['State'] == 'available':
                orph_ebs.append(flat_response[r])
    except Exception as e:
        print('failed to filter for orphan volumes' + str(e))

    '''read in all the snapshots for the region and store in global variable
    for use in snap report'''
    p = Paginator(vols_in_region, 150)
    for i in p:
        try:
            response = ec2_conn.describe_snapshots(Filters=[
                {
                    'Name': 'volume-id',
                    'Values': i,
                }
            ])['Snapshots']
            for i in response:
                non_orphan_snaps.append(i)

        except Exception as e:
            print('\nError locating EBS Snaps from' + str(region) +
                  ' ' + str(e))

    try:
        '''loop for parseing snapshot times + adding to snaptime dict'''
        now = datetime.datetime.utcnow()
        snaptime_dict = []
        for a in range(len(non_orphan_snaps)):
            last_24 = 0
            last_7 = 0
            last_30 = 0
            _30to365 = 0
            older = 0

            yesterday = (now - datetime.timedelta(days=1))
            lastweek = (now - datetime.timedelta(days=7))
            lastmonth = (now - datetime.timedelta(days=30))
            lastyear = (now - datetime.timedelta(days=365))

            snaptime = non_orphan_snaps[a]['StartTime']
            snaptime = snaptime.replace(tzinfo=None)

            if snaptime > yesterday:
                last_24 += 1
            if (snaptime > lastweek) and (snaptime < yesterday):
                last_7 += 1
            if (snaptime > lastmonth) and (snaptime < lastweek):
                last_30 += 1
            if snaptime < lastmonth and (snaptime < lastyear):
                _30to365 += 1
            if snaptime < lastyear:
                older += 1

            snaptime_dict.append({
                'VolumeId': non_orphan_snaps[a]['VolumeId'],
                'last_24': last_24,
                'last_7': last_7,
                'last_30': last_30,
                '_30to365': _30to365,
                'older': older
            })

    except Exception as e:
        print('Failed to parse snapshot time data\n' + str(e))

    try:
        '''write non-orphaned ebs data to data structure'''
        for i in range(len(active_ebs)):
            last_24 = 0
            last_7 = 0
            last_30 = 0
            _30to365 = 0
            older = 0
            for s in range(len(snaptime_dict)):
                if snaptime_dict[s]['VolumeId'] == active_ebs[i]['VolumeId']:
                    last_24 += snaptime_dict[s]['last_24']
                    last_7 += snaptime_dict[s]['last_7']
                    last_30 += snaptime_dict[s]['last_30']
                    _30to365 += snaptime_dict[s]['_30to365']
                    older += snaptime_dict[s]['older']

            '''encrypted true/false'''
            encryption = ''
            if active_ebs[i]['Encrypted']:
                encryption = 'Yes'
            else:
                encryption = 'No'

            active_dict.append({
                'VolumeId': active_ebs[i]['VolumeId'],
                'Size (GB)': active_ebs[i]['Size'],
                'Region': str(region),
                'Disk Type': active_ebs[i]['VolumeType'],
                'Encrypted?': encryption,
                'Snaps -24 Hours': last_24,
                'Snaps -7 Days': last_7,
                'Snaps -30 Days': last_30,
                'Snaps 30-365 Days': _30to365,
                'Snaps older than 1yr': older,
                'Total Snaps': (last_24 + last_7 + last_30 +
                                _30to365 + older)
            })

    except Exception as e:
        print('Failed to  write non-orphan ebs data to dict\n' + str(e))

    try:
        '''write orphaned ebs data to data structure'''
        for o in range(len(orph_ebs)):
            last_24 = 0
            last_7 = 0
            last_30 = 0
            _30to365 = 0
            older = 0
            for s in range(len(snaptime_dict)):
                if snaptime_dict[s]['VolumeId'] == active_ebs[i]['VolumeId']:
                    last_24 += snaptime_dict[s]['last_24']
                    last_7 += snaptime_dict[s]['last_7']
                    last_30 += snaptime_dict[s]['last_30']
                    _30to365 += snaptime_dict[s]['_30to365']
                    older += snaptime_dict[s]['older']

            '''encrypted true/false'''
            encryption = ''
            if orph_ebs[o]['Encrypted']:
                encryption = 'Yes'
            else:
                encryption = 'No'

            orph_dict.append({
                'VolumeId': orph_ebs[o]['VolumeId'],
                'Size (GB)': orph_ebs[o]['Size'],
                'Region': str(region),
                'Disk Type': orph_ebs[o]['VolumeType'],
                'Encrypted?': encryption,
                'Snaps -24 Hours': last_24,
                'Snaps -7 Days': last_7,
                'Snaps -30 Days': last_30,
                'Snaps 30-365 Days': _30to365,
                'Snaps older than 1yr': older,
                'Total Snaps': (last_24 + last_7 + last_30 +
                                _30to365 + older)
            })

    except Exception as e:
        print('Failed to write orphan ebs data to dict\n' + str(e))

    return active_dict, orph_dict


def ebs_snap_report(ec2_conn, snap_response, region):
    dict = []
    '''use pre-fetched snap data to write out python dict'''
    try:
        for i in range(len(snap_response)):
            '''check if snap is older than 60 days'''
            now = datetime.datetime.utcnow()
            snaptime = snap_response[i]['StartTime']
            snaptime = snaptime.replace(tzinfo=None)

            if snaptime <= (now - datetime.timedelta(days=60)):
                age = True
            else:
                age = False

            snapdate = snaptime.strftime("%m/%d/%Y")
            snaphours = snaptime.strftime("%H:%M")
            dict.append({
                'SnapshotId': snap_response[i]['SnapshotId'],
                'Source Vol ID': snap_response[i]['VolumeId'],
                'Source Vol Size (GB)': snap_response[i]['VolumeSize'],
                'Region': str(region),
                'Snap Date': snapdate,
                'Snap Time': snaphours,
                '60+ Days Old': age
            })
    except Exception as e:
        print('''\nError extracting snap report data from global
        snap variable''' + str(region) + ' ' + str(e))

    return dict


def orphan_ebs_snap_report(ec2_conn, region):
    '''get snaps NOT ASSOCIATED with vols_in_region'''
    dict = []
    orphans = []
    response = []
    flat_response = []

    '''fetch ALL snapshot information with paginator'''
    print('\nFetching Orphan EBS Snap Data from ' + str(region) + '...')

    pag = ec2_conn.get_paginator('describe_snapshots')
    for page in pag.paginate(OwnerIds=['self'],
                             PaginationConfig={'PageSize': 150}):
        for i in page['Snapshots']:
            response.append(i)

    '''flatten reponse structure'''
    for d in range(len(response)):
        flat_response.append(response[d])

    '''filter for orphans'''
    for s in flat_response:
        if s['VolumeId'] not in vols_in_region:
            orphans.append(s)

    '''parse individual snapshots and add to dict'''
    try:
        for i in range(len(orphans)):
            '''check if snap is older than 60 days'''
            now = datetime.datetime.utcnow()
            snaptime = orphans[i]['StartTime']
            snaptime = snaptime.replace(tzinfo=None)

            if snaptime <= (now - datetime.timedelta(days=60)):
                age = True
            else:
                age = False

            snapdate = snaptime.strftime("%m/%d/%Y")
            snaphours = snaptime.strftime("%H:%M")
            dict.append({
                'SnapshotId': orphans[i]['SnapshotId'],
                'Source Vol ID': orphans[i]['VolumeId'],
                'Source Vol Size (GB)': orphans[i]['VolumeSize'],
                'Region': str(region),
                'Encrypted?': orphans[i]['Encrypted'],
                'Snap Date': snapdate,
                'Snap Time': snaphours,
                '60+ Days Old': age

            })
    except Exception as e:
        print('\nError locating Orphan EBS Snaps from ' +
              str(region) + ':\n' + str(e))
    return dict


def analyze_engine_info(engine, engine_version, engine_mode):

    if engine_mode == "serverless":
        return False, False

    engine_version_parsed = engine_version.split(".")
    needs_upgrade = False

    try:
        if engine == "aurora" or engine == "aurora-mysql":
            major_version = (int(engine_version_parsed[0]), int(
                engine_version_parsed[1]))
            if major_version[0] != 5 or major_version[1] < 6:
                return False, False

            if major_version[1] == 6:
                needs_upgrade = engine_version < "5.6.mysql_aurora.1.19.2"
            elif major_version[1] == 7:
                needs_upgrade = engine_version < "5.7.mysql_aurora.2.04.4"

        elif engine == "mysql":
            major_version = (int(engine_version_parsed[0]), int(
                engine_version_parsed[1]))

            if major_version[0] == 5 and major_version[1] < 6:
                return False, False

            if major_version[1] == 6:
                needs_upgrade = engine_version < "5.6.40"
            elif major_version[1] == 7:
                needs_upgrade = engine_version < "5.7.24"
            elif major_version[1] == 8:
                needs_upgrade = engine_version < "8.0.13"

        elif engine == "postgres" or engine == "aurora-postgresql":
            major_version = (int(engine_version_parsed[0]), int(
                engine_version_parsed[1]))
            upgrade_targets = [((9, 4), 21), ((9, 5), 16),
                               ((9, 6), 12), ((10, 0), 7), ((11, 0), 2)]

            if major_version[0] > 11:
                return False, False

            elif major_version[0] == 9:
                for upgrade_target in upgrade_targets:
                    # ex: 9.6.12
                    if major_version[1] == upgrade_target[0][1] and int(engine_version_parsed[2]) < upgrade_target[1]:
                        needs_upgrade = True
                        break
            else:
                for upgrade_target in upgrade_targets:
                    # ex: 10.4
                    if major_version[0] == upgrade_target[0][0] and major_version[1] < upgrade_target[1]:
                        needs_upgrade = True
                        break
        else:
            return False, False
    except Exception:
        return False, False

    return True, needs_upgrade


def rds_inst_report(rds_conn, region):
    dict = []
    response_dict = []
    response = []
    

    '''create paginator and build rds_instance reponse dict'''
    print('\nFetching RDS instance Data from ' + str(region) + '...')
    pag = rds_conn.get_paginator('describe_db_instances')
    for page in pag.paginate(PaginationConfig={'PageSize': 99}):
        for i in page['DBInstances']:
            response.append(i)
    for d in range(len(response)):
        response_dict.append(response[d])

    print('\nFound ' + str(len(response_dict)) + ' RDS Instances in ' +
          str(region))

    '''write RDS data to data structure'''
    for i in range(len(response_dict)):
        dbsize = None
        engine = response_dict[i]['Engine']
        engine_version = response_dict[i]['EngineVersion']

        is_granular_backup_supported, needs_internal_upgrade \
            = analyze_engine_info(engine, engine_version, "")

        try:
            if engine == 'aurora-postgresql' or engine == 'aurora' or engine == 'aurora-mysql':
                dbsize = None
            else:
                dbsize = response_dict[i]['AllocatedStorage']

        except Exception as e:
            print('''\nError retreiving aurora snapshot for ''' +
                  str(response_dict[i]['DBInstanceIdentifier']) +
                  '\n' + str(e))

        '''perform aurora name translations'''
        if engine == 'aurora':
            engine = 'aurora-mysql'

        dict.append({
            'DB Instance Identifier': response_dict[i]['DBInstanceIdentifier'],
            'DB Provisioned Capacity (GB)': dbsize,
            'Region': str(region),
            'Status': response_dict[i]['DBInstanceStatus'],
            'Backup Retention Days': response_dict[i]['BackupRetentionPeriod'],
            'Engine': engine,
            'EngineVersion': engine_version,
            'Master User': response_dict[i]['MasterUsername'],
            'IsGranularBackupSupported': str(is_granular_backup_supported),
            'NeedsInternalUpgrade': str(needs_internal_upgrade),
            'DBInstanceStatus': response_dict[i]['DBInstanceStatus'],
            'Instance Type': response_dict[i]['DBInstanceClass']
        })

    return dict


def rds_clusters_report(rds_conn, region):
    dict = []
    response_dict = []
    response = []

    '''create paginator and build rds_instance reponse dict'''
    print('\nFetching RDS cluster Data from ' + str(region) + '...')

    pag = rds_conn.get_paginator('describe_db_clusters')
    for page in pag.paginate(PaginationConfig={'PageSize': 100}):
        for i in page['DBClusters']:
            response.append(i)
    for d in range(len(response)):
        response_dict.append(response[d])

    print('\nFound ' + str(len(response_dict)) + ' RDS Clusters in ' +
          str(region))

    '''write RDS data to data structure'''
    for i in range(len(response_dict)):
        allocatedGB = None
        engine = response_dict[i]['Engine']
        engine_version = response_dict[i]['EngineVersion']
        engine_mode = response_dict[i]['EngineMode']

        is_granular_backup_supported, needs_internal_upgrade \
            = analyze_engine_info(engine, engine_version, engine_mode)
        try:
            allocatedGB = get_aurora_allocated(
                response_dict[i]['DBClusterIdentifier'], rds_conn)
        except Exception as e:
            print('failed to retreived DB size for' +
                  str(response_dict[i]['DBClusterIdentifier']) + '\n' + str(e))

        '''perform aurora name translations'''
        if engine == 'aurora':
            engine = 'aurora-mysql'

        dict.append({
            'DB Cluster Identifier': response_dict[i]['DBClusterIdentifier'],
            'AllocatedStorage (GB)': allocatedGB,
            'Region': str(region),
            'Status': response_dict[i]['Status'],
            'Backup Retention Days': response_dict[i]['BackupRetentionPeriod'],
            'Engine': engine,
            'EngineVersion': engine_version,
            'EngineMode': engine_mode,
            'IsGranularBackupSupported': str(is_granular_backup_supported),
            'NeedsInternalUpgrade': str(needs_internal_upgrade),
            'MultiAZ': response_dict[i]['MultiAZ'],
            'CrossAccountClone': response_dict[i]['CrossAccountClone']
        })

    return dict


def connect_service(key_id, secret_key, region, service):
    '''connect client'''
    print('\nAttmpting connection to ' + service + ' in ' + str(region))
    try:
        client = boto3.client(
            service,
            aws_access_key_id=key_id,
            aws_secret_access_key=secret_key,
            region_name=region
        )
        print('\nSuccessful connection to ' + service + ' in ' + str(region))
        return client

    except ClientError as e:
        print('\nFailed to connect to ' + service + ' in ' + str(region)
              + str(e))


def dataframer(dict, index_name):
    try:
        df = pd.DataFrame(dict)
        df.set_index(index_name, inplace=True)
    except Exception as e:
        print('\nError adding ' + str(dict) + ' to dataframe ' + str(e))
        df = pd.DataFrame(columns=[str(index_name)])
    return df


def get_bucket_list(region, s3_conn):
    s3_data = s3_conn.list_buckets()
    s3_bucket_list = []
    for i in range(len(s3_data['Buckets'])):
        s3_bucket_list.append(s3_data['Buckets'][i]['Name'])

    return s3_bucket_list


def get_aurora_allocated(clusterID, rds_conn):
    response = rds_conn.describe_db_cluster_snapshots(
        DBClusterIdentifier=clusterID,
        SnapshotType='automated',
        MaxRecords=100,
    )
    return response['DBClusterSnapshots'][0]['AllocatedStorage']


class Paginator:
    def __init__(self, iterable, page_len=3):
        self.iterable = iterable
        self.page_len = page_len

    def __iter__(self):
        page = []
        for i in self.iterable:
            page.append(i)
            if len(page) == self.page_len:
                yield page
                page = []
        if page:
            yield page


if __name__ == '__main__':
    cli()
